/**
 * Copyright (c) 2017, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * WSO2 Inc. licenses this file to you under the Apache License,
 * Version 2.0 (the "License"); you may not use this file except
 * in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 *
 */
import _ from 'lodash';

let join;
const tab = '    ';

function times(n, f) {
    let s = '';
    for (let i = 0; i < n; i++) {
        s += f();
    }
    return s;
}

export default function getSourceOf(node, pretty = false, l = 0, replaceLambda) {
    if (!node) {
        return '';
    }
    let i = 0;
    const ws = node.ws ? node.ws.map(wsObj => wsObj.ws) : [];
    const shouldIndent = pretty || !ws.length;

    /**
     * White space generator function,
     * @param {string?} defaultWS
     * @return {string}
     */
    function w(defaultWS) {
        const wsI = ws[i++];
        // Check if whitespces have comments
        const hasComment = (wsI !== undefined) && wsI.trim().length > 0;
        if (hasComment || (!shouldIndent && wsI !== undefined)) {
            return wsI;
        }
        return defaultWS || '';
    }

    function a(afterWS) {
        if (shouldIndent) {
            return afterWS || '';
        }
        return '';
    }

    /* eslint-disable no-unused-vars */
    const b = a;

    function indent() {
        ++l;
        return '';
    }

    function outdent() {
        --l;
        if (shouldIndent) {
            if (node.documentationText) {
                const indent = _.last(node.documentationText.split('\n'));
                if (indent === _.repeat(tab, l)) {
                    // if documentation text already contains the correct dent
                    return '';
                }
            }
            return '\n' + _.repeat(tab, l);
        }
        return '';
    }

    function dent() {
        if (shouldIndent) {
            return '\n' + _.repeat(tab, l);
        }
        return '';
    }

    /* eslint-enable no-unused-vars */

    if (replaceLambda && node.kind === 'Lambda') {
        return '$ function LAMBDA $';
    }
    // if this is a primitive value, return value directly
    if (Object(node) !== node) {
        return node;
    }

    switch (node.kind) {
        case 'CompilationUnit':
            return join(node.topLevelNodes, pretty, replaceLambda, l, w) + w();
        case 'MarkdownDocumentation':
            let docString = "";

            for (let i = 0; i < node.ws.length; i++) {
                docString += node.ws[i].ws + node.ws[i].text;
                docString += node.documentationLines[i].text;
            }

            if (node.parameters) {
                for (let i = 0; i < node.parameters.length; i++) {
                    let parameter = node.parameters[i];
                    for (let j = 0; j < parameter.ws.length; j++) {
                        docString += parameter.ws[j].ws + parameter.ws[j].text;
                    }
                }
            }

            if (node.returnParameter) {
                for (let i = 0; i < node.returnParameter.ws.length; i++) {
                    docString += node.returnParameter.ws[i].ws + node.returnParameter.ws[i].text;
                }
            }

            return docString;
        /* eslint-disable max-len */
        // auto gen start
// auto-gen-code
        // auto gen end
        /* eslint-enable max-len */

        default:
            console.error('no source gen for ' + node.kind);
            return '';

    }
}

/**
 * Joins sources of a array of nodes with given delimiters.
 *
 * @private
 * @param {Node[]} arr Nodes to be joined.
 * @param {boolean} pretty
 * @param {number} l indent level.
 * @param {function(number): string} wsFunc White space generator function.
 * @param defaultWS
 * @param {string} separator
 * @param {boolean} suffixLast
 * @param {boolean} replaceLambda
 * @return {string}
 */
join = function (arr, pretty, replaceLambda, l, wsFunc, defaultWS, separator, suffixLast = false) {
    let str = '';
    for (let i = 0; i < arr.length; i++) {
        const node = arr[i];
        if (node.kind === 'Identifier') {
            str += wsFunc(defaultWS);
        }
        str += getSourceOf(node, pretty, l, replaceLambda);
        if (separator && (suffixLast || i !== arr.length - 1)) {
            str += wsFunc(defaultWS) + separator;
        }
    }

    return str;
};
